/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2011 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on May 10, 2011, 11:40 AM
 */

#include <fparser.h>
#include <math.h>

#include "boost/numeric/mtl/mtl.hpp"
#include "common/u2d_Math.h"
#include "exceptions.h"


using namespace mtl;


/**
 * This is an evaluator for 2D scalar fields. The field should be given in a
 * parametric style, so with a formula for x coordinate and another for y
 * coordinate.
 * @param points a matrix containing the coordinates of the points where
 * to evaluate the functions. The matrix is assumed this form:
 *         node 1  node 2 ...  node 3  node n
 *  rr = | 45.756  34.43  ...  0.3459  0.3428 |  coord x
 *       | 765.32  45.32  ...  4.6426  0.6583 |  coord y
 * @param formulaX Formula for X coordinate.
 * @param formulaY Formula for Y coordinate
 * @return A matrix with 2 rows, containing x and y coordinates, and of the same
 * number of columns as the given points matrix.
 */
dense2D<double> u2d_Math::FieldEvaluator(dense2D<double> points, string formulaX, string formulaY)
throw (u2d_Size_Missmatch_Exception &) {
    std::cout << "--== Evaluating expression \'(" << formulaX << "; " << formulaY << ")\' ==--" << std::endl;
    std::cout << "Points vector size is: " << num_rows(points) << "x" << num_cols(points) << std::endl;
    FunctionParser fpX;
    FunctionParser fpY;
    dense2D<double> result(2, num_cols(points));
    //result = 0.0; Initialization not needed.
    if (num_rows(points) == 2) {
        fpX.Parse(formulaX.c_str(), "x,y");
        fpY.Parse(formulaY.c_str(), "x,y");
    } else {
        throw u2d_Size_Missmatch_Exception("Number of matrix rows containing coordinates is not 2");
    }
    for(size_t i = 0; i < num_cols(points); i++) {
        double values[2];
        values[0] = points(0, i);
        values[1] = points(1, i);
        result(0, i) = fpX.Eval(values);
        result(1, i) = fpY.Eval(values);
        if(fpX.EvalError() != 0 || fpY.EvalError() != 0)  {
            std::stringstream msg;
            msg << "Possible error in evaluation, fparser error number: " << fpX.EvalError() << " or " << fpY.EvalError() << std::endl;
            throw u2d_Size_Missmatch_Exception(msg.str());
        }
    }
    std::cout << "Evaluated a field of size: (" << num_rows(result) << "x" << num_cols(result) << ")" << std::endl;
    std::cout << "--== Done ==--" << std::endl;
    return result;

}

/**
 * Evaluates the given formula in the given set of points. The formula must have
 * a number of variables that is the same number of rows as the given set of
 * points. The variables should be named x, y in 2D and x, y, z in 3D. The
 * returned vector is in the same order as the points passed.
 *
 * @param points a matrix containing the coordinated of the points where
 * to evaluate the function. The matrix is assumed this form:
 *         node 1  node 2 ...  node 3  node n
 *  rr = | 45.756  34.43  ...  0.3459  0.3428 |  coord x
 *       | 765.32  45.32  ...  4.6426  0.6583 |  coord y
 * @param formula The formula to evaluate with variables named x, y and z. Where
 * x is the first row of the given points matrix, y is the second and z the third.
 * @return A vector with the function evaluated in all the given points.
 */
dense_vector<double> u2d_Math::ScalarEvaluator(dense2D<double> points, string formula)
throw (u2d_Size_Missmatch_Exception &) {
    std::cout << "--== Evaluating expression \'" << formula << "\' ==--" << std::endl;
    std::cout << "Points vector size is: " << num_rows(points) << "x" << num_cols(points) << std::endl;
    FunctionParser fp;
    dense_vector<double> result(num_cols(points), 0.0);
    if (num_rows(points) == 1) {
        fp.Parse(formula.c_str(), "x");
    } else if (num_rows(points) == 2) {
        fp.Parse(formula.c_str(), "x,y");
    } else if (num_rows(points) == 3) {
        fp.Parse(formula.c_str(), "x,y,z");
    } else {
        throw u2d_Size_Missmatch_Exception("Number of matrix rows containing coordinates is not 1, 2 or 3");
    }
    for(size_t i = 0; i < num_cols(points); i++) {
        double values[3];
        values[0] = points(0, i);
        if(num_rows(points) >= 2) values[1] = points(1, i);
        if(num_rows(points) >= 3) values[2] = points(2, i);
        result(i) = fp.Eval(values);
        if(fp.EvalError() != 0)  {
            std::stringstream msg;
            msg << "Possible error in evaluation, fparser error number: " << fp.EvalError() << std::endl;
            throw u2d_Size_Missmatch_Exception(msg.str());
        }
    }
    std::cout << "Evaluated a vector of: " << size(result) << " points" << std::endl;
    std::cout << "--== Done ==--" << std::endl;
    return result;
}

/**
 * 
 * @param color A pointer to a return value array of size at least of
 * 3*sizeof(float) that will hold calculated color values. Values will be
 * between 0.0 and 1.0, where 1.0 correspond to 255.
 * @param value
 * @param minValue
 * @param maxValue
 */
void u2d_Math::LinearRGB(float *color, float value, float minValue, float maxValue)
throw (u2d_Math_Exception &) {
    if((value > maxValue) || (value < minValue)) {
        throw u2d_Math_Exception("Given value is out of range for RGB color linearization.");
    }
    /* The resulting color */
    float r = 0;
    float g = 0;
    float b = 0;
    
    /* The delta between max and min input */
    float inputDelta = maxValue - minValue;
    /* The normalized value, 0->1 */
    float normalizedValue = (value - minValue) / inputDelta;
    /* The color normalized value, we have 4 areas of color we use for
     * linearization of RGB space, each area has a size of 256, so the max
     * value is 256 * 4 */
    float colorNormalizedValue = normalizedValue * 255.0 * 4.0;
    
    /* Linearization algorithm */
    if(colorNormalizedValue < 255 ) {
        r = 255;
        g = fmod(colorNormalizedValue, 255.0);
        b = 0;
    } else if((colorNormalizedValue >= 255) && (colorNormalizedValue < 510 )) {
        r = 255 - fmod(colorNormalizedValue, 255.0);
        g = 255;
        b = 0;
    } else if((colorNormalizedValue >= 510) && (colorNormalizedValue < 765 )) {
        r = 0;
        g = 255;
        b = fmod(colorNormalizedValue, 255.0);
    } else if((colorNormalizedValue >= 765) && (colorNormalizedValue < 1020 )) {
        r = 0;
        g = 255 - fmod(colorNormalizedValue, 255.0);
        b = 255;
    } else if(colorNormalizedValue = 1020 ) {
        r = 0;
        g = 0;
        b = 255;
    } else {
        throw u2d_Math_Exception("Cannot linearize to color value. Possible algorithm error.");
    }
    /* Set the colors to return */
    color[0] = r / 255.0;
    color[1] = g / 255.0;
    color[2] = b / 255.0;
}
